/****************************************************************/
/*  
     ***   ***  ***  *** ***     *** ***   ******  ****  ***
     ***   ***  ***  *** ***     *** **   ***  *** ***** ***
     ***   ***  ***  *** ***     *****    ******** *** *****
      *** ***   ***  *** ******* *** **   ***  *** ***  ****
       *****     *****   ******* *** ***  ***  *** ***   ***

	 Vulkan Graphics API
	 Introductory Tutorials

   Minimal working examples (MWE) to build upon and 
   modify to gain an understanding of the API

/*
/****************************************************************/
/*
   About:
   Debug drawing class

*/
/****************************************************************/

#if 1

#pragma once

#include <vector>
using namespace std;

#include "helpers.h" // simple vector/matrix classes


// this is if we do not declared prototypes, 
// and want to load dynamically!
//#define VK_NO_PROTOTYPES            
                                      
#include "vulkan.h"



/**************************************************************/
/*
   Create simple model meshes (plane, quad, cube) - for debugging/testing

   Model
*/
/**************************************************************/

#define VERTEX_BUFFER_BIND_ID 0

#define MAX_MESH_VERTICES     1000
#define MAX_MESH_INDICES      1000


class DebugMesh
{
public:
  static 
  void Create(VkDevice         device,
              VkPhysicalDevice physicalDevice,
              int width,
              int height,
              VkRenderPass            renderPass)
  {
    vbuf               = VK_NULL_HANDLE;
    vmem               = VK_NULL_HANDLE;
    vsize              = 0;
	  
    ibuf               = VK_NULL_HANDLE;
    imem               = VK_NULL_HANDLE;
    isize              = 0;

    ubuf               = VK_NULL_HANDLE;
    umem               = VK_NULL_HANDLE;
    usize              = 0;
    
    indexCount         = 0;
    dim                = Vector3(0,0,0);
    memset(&verticeDescription, 0, sizeof(verticeDescription));

		CreateMesh(device, physicalDevice);

		SetupVertexDescriptions();

    SetupDescriptors(device,
                     physicalDevice,
                     &descriptorSet,
                     &descriptorSetLayout);

    SetupPipeline(device,
                  width,
                  height,
                  descriptorSetLayout,
                  vertShaderModule,
                  fragShaderModule,
                  renderPass,
                  &pipeline,
                  &pipelineLayout);
  }

  static 
  void Release()
  {
  }

  static
  void Present(VkDevice         device,
               VkPhysicalDevice physicalDevice,
               uint32_t         nextImageIdx,
               VkFramebuffer*   frameBuffers,
               int              width,
               int              height,
               VkCommandBuffer  commandBuffer,
               VkRenderPass     renderPass)
  {
    if ( vertexBuffer.size() <=0 ) return;
    if ( indexBuffer.size() <=0  ) return;
    
    UpdateUniformBuffers(device,
                         physicalDevice);

    UpdateMeshBuffers(device,
                      physicalDevice);

    if ( indexCount <=0          ) return;


    UpdateCommandBuffer(device,
                        physicalDevice,
                        nextImageIdx,
                        frameBuffers,
                        width,
                        height,
                        commandBuffer,
                        renderPass,
                        descriptorSet,
                        pipelineLayout,
                        pipeline);
    // draw mesh 
    {
    VkFence renderFence;
    VkFenceCreateInfo fenceCreateInfo  = {};
    fenceCreateInfo.sType              = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO;
    vkCreateFence( device, 
                   &fenceCreateInfo, 
                   NULL, 
                   &renderFence );

    VkPipelineStageFlags waitStageMash = { VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT };

    VkSubmitInfo submitInfo            = {};
    submitInfo.sType                   = VK_STRUCTURE_TYPE_SUBMIT_INFO;
    submitInfo.waitSemaphoreCount      = 0;
    submitInfo.pWaitSemaphores         = VK_NULL_HANDLE;
    submitInfo.pWaitDstStageMask       = &waitStageMash;
    submitInfo.commandBufferCount      = 1;
    submitInfo.pCommandBuffers         = &commandBuffer;
    submitInfo.signalSemaphoreCount    = 0;
    submitInfo.pSignalSemaphores       = VK_NULL_HANDLE;

    VkQueue presentQueue = 0;
    vkGetDeviceQueue( device, 
                    0, 
                    0, 
                    &presentQueue );

    // Submit command buffer (drawing)
    vkQueueSubmit(  presentQueue, 
            1, 
            &submitInfo, 
            renderFence );

    // Wait for the the command/queue to finish
    vkWaitForFences( device, 1, &renderFence, VK_TRUE, UINT64_MAX );
    vkDestroyFence( device, renderFence, NULL );
    }

    vertexBuffer.clear();
    indexBuffer.clear();
  }

  static 
  void SetModelViewProjection(const Matrix4& m,
                              const Matrix4& v,
                              const Matrix4& p)
  {
    modelMatrix      = m;
    viewMatrix       = v;
    projectionMatrix = p;
  }

  static 
  void AddDrawLine(const Vector3& p0, 
                   const Vector3& p1,
                   const Vector3& color = Vector3(1,0,0) )
  {
    stVertex v0 = {p0, color};
    stVertex v1 = {p1, color};

    vertexBuffer.push_back( v0 );
    vertexBuffer.push_back( v1 );
    vertexBuffer.push_back( v1 );
    DBG_ASSERT(vertexBuffer.size() < MAX_MESH_VERTICES );

    indexBuffer.push_back( vertexBuffer.size()-3 );
    indexBuffer.push_back( vertexBuffer.size()-2 );
    indexBuffer.push_back( vertexBuffer.size()-1 );
    DBG_ASSERT(indexBuffer.size() < MAX_MESH_VERTICES );
  }

  static 
  void AddDrawCross(const Vector3& p0, 
                    const float length = 1.0f,
                    const Vector3& color = Vector3(1,0,0) )
  {
    AddDrawLine( p0-Vector3(1,0,0)*length, p0+Vector3(1,0,0)*length, color );
    AddDrawLine( p0-Vector3(0,1,0)*length, p0+Vector3(0,1,0)*length, color );
    AddDrawLine( p0-Vector3(0,0,1)*length, p0+Vector3(0,0,1)*length, color );

  }

private:
	struct stVertex 
	{
		Vector3 pos;
		Vector3 col;
	};

	struct stVerticeDescription
	{
		VkPipelineVertexInputStateCreateInfo           inputState;
		std::vector<VkVertexInputBindingDescription>   bindingDescriptions;
		std::vector<VkVertexInputAttributeDescription> attributeDescriptions;
	};

  // Structure must match the structure defined uniform in shader
  struct stUniform 
  {
    Matrix4 model;
    Matrix4 view;
    Matrix4 projection;
  };

	static VkBuffer			         vbuf;
	static VkDeviceMemory        vmem;
	static size_t				         vsize;

	static VkBuffer			         ibuf;
	static VkDeviceMemory	       imem;
	static size_t				         isize;

  static VkBuffer			         ubuf;
	static VkDeviceMemory	       umem;
	static size_t				         usize;

	static uint32_t			         indexCount;
	static Vector3				       dim;
	static stVerticeDescription  verticeDescription;

  static vector<stVertex>      vertexBuffer;
  static vector<uint32_t>      indexBuffer;
  static Matrix4               modelMatrix;
  static Matrix4               viewMatrix;
  static Matrix4               projectionMatrix;
  static VkShaderModule        vertShaderModule;
  static VkShaderModule        fragShaderModule;


  static VkPipelineLayout      pipelineLayout;
  static VkPipeline            pipeline;
  static VkDescriptorSet       descriptorSet;
  static VkDescriptorSetLayout descriptorSetLayout;

  
  static
  void UpdateCommandBuffer(VkDevice         device,
                           VkPhysicalDevice physicalDevice,
                           uint32_t         nextImageIdx,
                           VkFramebuffer*   frameBuffers,
                           int              width,
                           int              height,
                           VkCommandBuffer  commandBuffer,
                           VkRenderPass     renderPass,
                           VkDescriptorSet  descriptorSet,
                           VkPipelineLayout pipelineLayout,
                           VkPipeline       pipeline)
  {
    
  // Build the command buffer

  VkCommandBufferBeginInfo beginInfo            = {};
	beginInfo.sType                               = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
  beginInfo.flags                               = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT;
	beginInfo.pNext                               = NULL;

  vkBeginCommandBuffer(commandBuffer, 
                       &beginInfo );


  // bind the graphics pipeline to the command buffer. 
  // Any vkDraw command afterwards is affected by this
  // pipeline.
  vkCmdBindPipeline( commandBuffer, 
                     VK_PIPELINE_BIND_POINT_GRAPHICS, 
                     pipeline );    

  // take care of dynamic state:
  VkViewport viewport = { 0, 0, (float)width, (float)height, 0, 1 };
  vkCmdSetViewport( commandBuffer, 0, 1, &viewport );

  VkRect2D scissor = { 0, 0, width, height };
  vkCmdSetScissor( commandBuffer, 0, 1, &scissor);

  // bind our shader parameters:
  vkCmdBindDescriptorSets( commandBuffer, 
                           VK_PIPELINE_BIND_POINT_GRAPHICS, 
                           pipelineLayout, 
                           0, 
                           1, 
                           &descriptorSet, 
                           0, 
                           NULL );

  DBG_ASSERT(vertexBuffer.size()>0);
  DBG_ASSERT(indexBuffer.size()>0);

  VkDeviceSize offsets = { 0 };
	vkCmdBindVertexBuffers(commandBuffer, VERTEX_BUFFER_BIND_ID, 1, &vbuf, &offsets);
	vkCmdBindIndexBuffer(commandBuffer, ibuf, 0, VK_INDEX_TYPE_UINT32);
  vkCmdDrawIndexed(commandBuffer, indexCount, 1, 0, 0, 0);


  //vkCmdEndRenderPass( commandBuffer );

  vkEndCommandBuffer( commandBuffer );
  
  }// End UpdateCommandBuffer()
  


  static
	void SetupVertexDescriptions()
	{
		// Binding description
		verticeDescription.bindingDescriptions.resize(1);

		VkVertexInputBindingDescription vInputBindDescription = {};
		vInputBindDescription.binding                         = VERTEX_BUFFER_BIND_ID;
		vInputBindDescription.stride                          = sizeof(stVertex);
		vInputBindDescription.inputRate                       = VK_VERTEX_INPUT_RATE_VERTEX;
		verticeDescription.bindingDescriptions[0]             = vInputBindDescription;

		// Attribute descriptions
		verticeDescription.attributeDescriptions.resize(2);

		// Location 0 : Position
		verticeDescription.attributeDescriptions[0].location  = 0;
		verticeDescription.attributeDescriptions[0].binding   = VERTEX_BUFFER_BIND_ID;
		verticeDescription.attributeDescriptions[0].format    = VK_FORMAT_R32G32B32_SFLOAT;
		verticeDescription.attributeDescriptions[0].offset    = 0;

		// Location 1 : Color
		verticeDescription.attributeDescriptions[1].location  = 1;
		verticeDescription.attributeDescriptions[1].binding   = VERTEX_BUFFER_BIND_ID;
		verticeDescription.attributeDescriptions[1].format    = VK_FORMAT_R32G32B32_SFLOAT;
		verticeDescription.attributeDescriptions[1].offset    = sizeof(float) * 3;


		VkPipelineVertexInputStateCreateInfo pipelineVertexInputStateCreateInfo = {};
		pipelineVertexInputStateCreateInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO;
		pipelineVertexInputStateCreateInfo.pNext = NULL;

		verticeDescription.inputState                                 = pipelineVertexInputStateCreateInfo;
		verticeDescription.inputState.vertexBindingDescriptionCount   = verticeDescription.bindingDescriptions.size();
		verticeDescription.inputState.pVertexBindingDescriptions      = verticeDescription.bindingDescriptions.data();
		verticeDescription.inputState.vertexAttributeDescriptionCount = verticeDescription.attributeDescriptions.size();
		verticeDescription.inputState.pVertexAttributeDescriptions    = verticeDescription.attributeDescriptions.data();
	}// End SetupVertexDescriptions(..)


  static
	void CreateMesh(VkDevice         device,
                  VkPhysicalDevice physicalDevice)
	{		
		CreateBuffer(
			device,
      physicalDevice,
			VK_BUFFER_USAGE_VERTEX_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT,
			MAX_MESH_VERTICES * sizeof(stVertex), 
			nullptr, 
			&vbuf,
			&vmem);

		CreateBuffer(
			device,
      physicalDevice,
			VK_BUFFER_USAGE_INDEX_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT,
			MAX_MESH_INDICES * sizeof(uint32_t), 
			nullptr,
			&ibuf,
			&imem);

    CreateBuffer(
			device,
      physicalDevice,
			VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT,
			sizeof(stUniform),
			nullptr,
			&ubuf,
			&umem);

    LoadShader( device, ".\\debugmesh.vert.spv", &vertShaderModule);
    LoadShader( device, ".\\debugmesh.frag.spv", &fragShaderModule);

	}




  static
  void SetupDescriptors(VkDevice                device,
                      VkPhysicalDevice        physicalDevice,
                      VkDescriptorSet*        outDescriptorSet,
                      VkDescriptorSetLayout*  outDescriptorSetLayout)
{
  // Define and create our descriptors:
  VkDescriptorSetLayoutBinding bindings[1];

  // uniform buffer for our matrices:
  bindings[0].binding                                 = 0;
  bindings[0].descriptorType                          = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
  bindings[0].descriptorCount                         = 1;
  bindings[0].stageFlags                              = VK_SHADER_STAGE_VERTEX_BIT;
  bindings[0].pImmutableSamplers                      = NULL;



  VkDescriptorSetLayoutCreateInfo setLayoutCreateInfo = {};
  setLayoutCreateInfo.sType                           = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO;
  setLayoutCreateInfo.bindingCount                    = 1;
  setLayoutCreateInfo.pBindings                       = bindings;


  VkResult result = 
  vkCreateDescriptorSetLayout( device, 
                               &setLayoutCreateInfo, 
                               NULL, 
                               outDescriptorSetLayout );
  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to create DescriptorSetLayout." );

  // decriptor pool creation:
  VkDescriptorPoolSize uniformBufferPoolSize[1];
  uniformBufferPoolSize[0].type                       = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
  uniformBufferPoolSize[0].descriptorCount            = 1;


  VkDescriptorPoolCreateInfo poolCreateInfo           = {}; 
  poolCreateInfo.sType                                = VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO;
  poolCreateInfo.maxSets                              = 1;
  poolCreateInfo.poolSizeCount                        = 1;
  poolCreateInfo.pPoolSizes                           = uniformBufferPoolSize;


  VkDescriptorPool descriptorPool;
  result = vkCreateDescriptorPool( device, 
                                   &poolCreateInfo, 
                                   NULL, 
                                   &descriptorPool );
  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to create descriptor pool." );

  // allocate our descriptor from the pool:
  VkDescriptorSetAllocateInfo dsai                   = {};
  dsai.sType                                         = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
  dsai.descriptorPool                                = descriptorPool;
  dsai.descriptorSetCount                            = 1;
  dsai.pSetLayouts                                   = outDescriptorSetLayout;

  result = 
  vkAllocateDescriptorSets( device, 
                            &dsai, 
                            outDescriptorSet );

  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to allocate descriptor sets." );

  // When a set is allocated all values are undefined
  // and all descriptors are uninitialized. You must 
  // init all statically used bindings:
  VkDescriptorBufferInfo dbi                        = {};
  dbi.buffer                                        = ubuf;
  dbi.offset                                        = 0;
  dbi.range                                         = VK_WHOLE_SIZE;


  VkWriteDescriptorSet wd[1]                        = {};
  wd[0].sType                                       = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
  wd[0].dstSet                                      = *outDescriptorSet;
  wd[0].dstBinding                                  = 0;
  wd[0].dstArrayElement                             = 0;
  wd[0].descriptorCount                             = 1;
  wd[0].descriptorType                              = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
  wd[0].pImageInfo                                  = NULL;
  wd[0].pBufferInfo                                 = &dbi;
  wd[0].pTexelBufferView                            = NULL;

  vkUpdateDescriptorSets( device, 
                          1, 
                          wd, 
                          0, 
                          NULL );
}// End SetupDescriptors(..)


  static
  void SetupPipeline( VkDevice                device,
                      int                     width,
                      int                     height,
                      VkDescriptorSetLayout   descriptorSetLayout,
                      VkShaderModule          vertShaderModule,
                      VkShaderModule          fragShaderModule,
                      VkRenderPass            renderPass, 
                      VkPipeline*             outPipeline,
                      VkPipelineLayout*       outPipelineLayout)
{
  // Graphics Pipeline:
  VkPipelineLayoutCreateInfo layoutCreateInfo               = {};
  layoutCreateInfo.sType                                    = VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO;
  layoutCreateInfo.setLayoutCount                           = 1;
  layoutCreateInfo.pSetLayouts                              = &descriptorSetLayout;
  layoutCreateInfo.pushConstantRangeCount                   = 0;
  layoutCreateInfo.pPushConstantRanges                      = NULL;

  VkResult result = 
  vkCreatePipelineLayout( device, 
              &layoutCreateInfo, 
              NULL, 
              outPipelineLayout );
  DBG_ASSERT_VULKAN_MSG( result, "Failed to create pipeline layout." );



  // setup shader stages in the pipeline:
  VkPipelineShaderStageCreateInfo shaderStageCreateInfo[2]  = {};
  shaderStageCreateInfo[0].sType                            = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
  shaderStageCreateInfo[0].stage                            = VK_SHADER_STAGE_VERTEX_BIT;
  shaderStageCreateInfo[0].module                           = vertShaderModule;
  // shader entry point function name                       
  shaderStageCreateInfo[0].pName                            = "main";
  shaderStageCreateInfo[0].pSpecializationInfo              = NULL;

  shaderStageCreateInfo[1].sType                            = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
  shaderStageCreateInfo[1].stage                            = VK_SHADER_STAGE_FRAGMENT_BIT;
  shaderStageCreateInfo[1].module                           = fragShaderModule;
  // shader entry point function name                       
  shaderStageCreateInfo[1].pName                            = "main";
  shaderStageCreateInfo[1].pSpecializationInfo              = NULL;


  // vertex topology config:
  VkPipelineInputAssemblyStateCreateInfo inputAssemblyStateCreateInfo = {};
  inputAssemblyStateCreateInfo.sType                        = VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO;
  inputAssemblyStateCreateInfo.topology                     = VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST;
  inputAssemblyStateCreateInfo.primitiveRestartEnable       = VK_FALSE;

  // viewport config:
  VkViewport viewport                                       = {};
  viewport.x                                                = 0;
  viewport.y                                                = 0;
  viewport.width                                            = (float)width;
  viewport.height                                           = (float)height;
  viewport.minDepth                                         = 0;
  viewport.maxDepth                                         = 1;

  VkRect2D scissors                                         = {};
  VkOffset2D k                                              = {0,0};
  VkExtent2D m                                              = {width,height };
  scissors.offset                                           = k;
  scissors.extent                                           = m;

  VkPipelineViewportStateCreateInfo viewportState           = {};
  viewportState.sType                                       = VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO;
  viewportState.viewportCount                               = 1;
  viewportState.pViewports                                  = &viewport;
  viewportState.scissorCount                                = 1;
  viewportState.pScissors                                   = &scissors;

  // rasterization config:
  VkPipelineRasterizationStateCreateInfo rasterizationState = {};
  rasterizationState.sType                                  = VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO;
  rasterizationState.depthClampEnable                       = VK_FALSE;
  rasterizationState.rasterizerDiscardEnable                = VK_FALSE;
#if 0 // SOLID or WIREFRAME
  rasterizationState.polygonMode                            = VK_POLYGON_MODE_FILL;
#else
  rasterizationState.polygonMode                            = VK_POLYGON_MODE_LINE;
#endif
  rasterizationState.cullMode                               = VK_CULL_MODE_NONE;
  rasterizationState.frontFace                              = VK_FRONT_FACE_COUNTER_CLOCKWISE;
  rasterizationState.depthBiasEnable                        = VK_FALSE;
  rasterizationState.depthBiasConstantFactor                = 0;
  rasterizationState.depthBiasClamp                         = 0;
  rasterizationState.depthBiasSlopeFactor                   = 0;
  rasterizationState.lineWidth                              = 1;

  // sampling config:
  VkPipelineMultisampleStateCreateInfo multisampleState     = {};
  multisampleState.sType                                    = VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO;
  multisampleState.rasterizationSamples                     = VK_SAMPLE_COUNT_1_BIT;
  multisampleState.sampleShadingEnable                      = VK_FALSE;
  multisampleState.minSampleShading                         = 0;
  multisampleState.pSampleMask                              = NULL;
  multisampleState.alphaToCoverageEnable                    = VK_FALSE;
  multisampleState.alphaToOneEnable                         = VK_FALSE;



  // color blend config: (Actually off for tutorial)
  VkPipelineColorBlendAttachmentState colorBlendAttachmentState = {};
  colorBlendAttachmentState.blendEnable                     = VK_FALSE;
  colorBlendAttachmentState.srcColorBlendFactor             = VK_BLEND_FACTOR_SRC_COLOR;
  colorBlendAttachmentState.dstColorBlendFactor             = VK_BLEND_FACTOR_ONE_MINUS_DST_COLOR;
  colorBlendAttachmentState.colorBlendOp                    = VK_BLEND_OP_ADD;
  colorBlendAttachmentState.srcAlphaBlendFactor             = VK_BLEND_FACTOR_ZERO;
  colorBlendAttachmentState.dstAlphaBlendFactor             = VK_BLEND_FACTOR_ZERO;
  colorBlendAttachmentState.alphaBlendOp                    = VK_BLEND_OP_ADD;
  colorBlendAttachmentState.colorWriteMask                  = 0xf;

  VkPipelineColorBlendStateCreateInfo colorBlendState       = {};
  colorBlendState.sType                                     = VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO;
  colorBlendState.logicOpEnable                             = VK_FALSE;
  colorBlendState.logicOp                                   = VK_LOGIC_OP_CLEAR;
  colorBlendState.attachmentCount                           = 1;
  colorBlendState.pAttachments                              = &colorBlendAttachmentState;
  colorBlendState.blendConstants[0]                         = 0.0;
  colorBlendState.blendConstants[1]                         = 0.0;
  colorBlendState.blendConstants[2]                         = 0.0;
  colorBlendState.blendConstants[3]                         = 0.0;

  // configure dynamic state:
  VkDynamicState dynamicState[2]                            = { VK_DYNAMIC_STATE_VIEWPORT, VK_DYNAMIC_STATE_SCISSOR };
  VkPipelineDynamicStateCreateInfo dynamicStateCreateInfo   = {};
  dynamicStateCreateInfo.sType                              = VK_STRUCTURE_TYPE_PIPELINE_DYNAMIC_STATE_CREATE_INFO;
  dynamicStateCreateInfo.dynamicStateCount                  = 2;
  dynamicStateCreateInfo.pDynamicStates                     = dynamicState;

  // add depth buffer options to the pipeline
  VkPipelineDepthStencilStateCreateInfo depthStencil        = {};
  depthStencil.sType                                        = VK_STRUCTURE_TYPE_PIPELINE_DEPTH_STENCIL_STATE_CREATE_INFO;
  depthStencil.depthTestEnable                              = VK_TRUE;
  depthStencil.depthWriteEnable                             = VK_TRUE;
  depthStencil.depthCompareOp                               = VK_COMPARE_OP_LESS;
  depthStencil.depthBoundsTestEnable                        = VK_FALSE;
  depthStencil.minDepthBounds                               = 0.0f; // Optional
  depthStencil.maxDepthBounds                               = 1.0f; // Optional
  depthStencil.stencilTestEnable                            = VK_FALSE;


  // and finally, pipeline config and creation:
  VkGraphicsPipelineCreateInfo pipelineCreateInfo           = {};
  pipelineCreateInfo.sType                                  = VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO;
  pipelineCreateInfo.stageCount                             = 2;
  pipelineCreateInfo.pStages                                = shaderStageCreateInfo;
  pipelineCreateInfo.pVertexInputState                      = &verticeDescription.inputState;
  pipelineCreateInfo.pInputAssemblyState                    = &inputAssemblyStateCreateInfo;
  pipelineCreateInfo.pTessellationState                     = NULL;
  pipelineCreateInfo.pViewportState                         = &viewportState;
  pipelineCreateInfo.pRasterizationState                    = &rasterizationState;
  pipelineCreateInfo.pMultisampleState                      = &multisampleState;
  pipelineCreateInfo.pDepthStencilState                     = &depthStencil;
  pipelineCreateInfo.pColorBlendState                       = &colorBlendState;
  pipelineCreateInfo.pDynamicState                          = &dynamicStateCreateInfo;
  pipelineCreateInfo.layout                                 = *outPipelineLayout;
  pipelineCreateInfo.renderPass                             = renderPass;
  pipelineCreateInfo.subpass                                = 0;
  pipelineCreateInfo.basePipelineHandle                     = NULL;
  pipelineCreateInfo.basePipelineIndex                      = 0;

  result = 
  vkCreateGraphicsPipelines( device, 
                             VK_NULL_HANDLE, 
                             1, 
                             &pipelineCreateInfo, 
                             NULL, 
                             outPipeline );

  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to create graphics pipeline." );
  }

  static 
  void UpdateMeshBuffers(VkDevice              device,
                         VkPhysicalDevice      physicalDevice)
  {
      DBG_ASSERT(vertexBuffer.size()>0);
      DBG_ASSERT(indexBuffer.size()>0);

      int sizevb   = sizeof(stVertex) * vertexBuffer.size();
      void* vbdata = vertexBuffer.data();
      void *mapped;
			DBG_ASSERT_VULKAN(vkMapMemory(device, vmem, 0, sizevb, 0, &mapped));
			memcpy(mapped, vbdata, (size_t)sizevb);
			vkUnmapMemory(device, vmem);
      mapped = nullptr;


      int   sizeib = sizeof(stVertex) * indexBuffer.size();
      void* ibdata = indexBuffer.data();
			DBG_ASSERT_VULKAN(vkMapMemory(device, imem, 0, sizeib, 0, &mapped));
			memcpy(mapped, ibdata, (size_t)sizeib);
			vkUnmapMemory(device, imem);
      mapped = nullptr;



      indexCount = indexBuffer.size();
  }

  static 
  void UpdateUniformBuffers(VkDevice              device,
                            VkPhysicalDevice      physicalDevice)
  {
    // Update uniform buffers
    stUniform ubo;
    ubo.model      = (modelMatrix);
    ubo.view       = (viewMatrix);
    ubo.projection = (projectionMatrix);
    
	  uint8_t *pData;
	  DBG_ASSERT_VULKAN(vkMapMemory(device, umem, 0, sizeof(ubo), 0, (void **)&pData));
	  memcpy(pData, &ubo, sizeof(ubo));
	  vkUnmapMemory(device, umem);
  }



  static
	VkBool32 CreateBuffer(VkDevice              device,
                        VkPhysicalDevice      physicalDevice,
	                      VkBufferUsageFlags    usageFlags, 
	                      VkMemoryPropertyFlags memoryPropertyFlags, 
						            VkDeviceSize          size, 
						            void *                data, 
						            VkBuffer *            buffer, 
						            VkDeviceMemory *      memory)
	{
		VkMemoryRequirements memReqs;
		VkMemoryAllocateInfo memAlloc = {};
		memAlloc.sType           = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
		memAlloc.pNext           = NULL;
		memAlloc.allocationSize  = 0;
		memAlloc.memoryTypeIndex = 0;

		VkBufferCreateInfo bufferCreateInfo = {};
		bufferCreateInfo.sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
		bufferCreateInfo.pNext = NULL;
		bufferCreateInfo.usage = usageFlags;
		bufferCreateInfo.size  = size;
		bufferCreateInfo.flags = 0;


		DBG_ASSERT_VULKAN(vkCreateBuffer(device, &bufferCreateInfo, nullptr, buffer));

		vkGetBufferMemoryRequirements(device, *buffer, &memReqs);
		memAlloc.allocationSize  = memReqs.size;
		memAlloc.memoryTypeIndex = GetMemoryType(physicalDevice, memReqs.memoryTypeBits, memoryPropertyFlags);
		DBG_ASSERT_VULKAN(vkAllocateMemory(device, &memAlloc, nullptr, memory));
		if (data != nullptr)
		{
			void *mapped;
			DBG_ASSERT_VULKAN(vkMapMemory(device, *memory, 0, size, 0, &mapped));
			memcpy(mapped, data, (size_t)size);
			vkUnmapMemory(device, *memory);
		}
		DBG_ASSERT_VULKAN(vkBindBufferMemory(device, *buffer, *memory, 0));

		return true;
	}

  static
  uint32_t GetMemoryType(VkPhysicalDevice      physicalDevice,
                         uint32_t              typeBits, 
                         VkMemoryPropertyFlags properties)
  {
		// read the device memory properties
		VkPhysicalDeviceMemoryProperties memoryProperties;
		vkGetPhysicalDeviceMemoryProperties( physicalDevice, &memoryProperties );

    bool memTypeFound = false;
		for (uint32_t i = 0; i < memoryProperties.memoryTypeCount; i++)
		{
			if ((typeBits & 1) == 1)
			{
				if ((memoryProperties.memoryTypes[i].propertyFlags & properties) == properties)
				{
					if (memTypeFound)
					{
						memTypeFound = true;
					}
					return i;
				}
			}
			typeBits >>= 1;
		}

		if (!memTypeFound)
		{
			__debugbreak(); // Could not find a matching memory type
      return 0;
		}// End if
		return 1;
	}// End GetMemoryType(..)


}; // End class DebugMesh




#endif

