/****************************************************************/
/*  
     ***   ***  ***  *** ***     *** ***   ******  ****  ***
     ***   ***  ***  *** ***     *** **   ***  *** ***** ***
     ***   ***  ***  *** ***     *****    ******** *** *****
      *** ***   ***  *** ******* *** **   ***  *** ***  ****
       *****     *****   ******* *** ***  ***  *** ***   ***

	 Vulkan Graphics API
	 Introductory Tutorials

   Minimal working examples (MWE) to build upon and 
   modify to gain an understanding of the API

/*
/****************************************************************/
/*
     About:
	 helper.h set of helper functions to reduce the complexity
	 of the `demos' without introducing a complete framework.

	 Simple math classes, e.g., Vector, Matrix,

*/

#pragma once



// Windows version of the Vulkan API
#define VK_USE_PLATFORM_WIN32_KHR

// this is if we do not declared prototypes, 
// and want to load dynamically!
//#define VK_NO_PROTOTYPES            
                                      
#include "vulkan.h"




//Saving debug information to a log file/screen/..
inline 
void dprintf(const char *fmt, ...) 
{
	va_list parms;
	static char buf[2048] = {0};

	// Try to print in the allocated space.
	va_start(parms, fmt);
	vsprintf_s(buf, fmt, parms);
	va_end(parms);

	// Write the information out to a txt file
	#if 0
	FILE *fp = fopen("output.txt", "a+");
	fprintf(fp, "%s",  buf);
	fclose(fp);
	#endif

	// Output to the visual studio window
	OutputDebugStringA( buf );

}// End dprintf(..)



// Debug defines (custom asserts)
#if defined(_WIN32)
  #define DBG_ASSERT(f) { if(!(f)){ __debugbreak(); }; }
#else
  #define DBG_ASSERT(f) { #error(platform assert todo) }
#endif

#define DBG_VALID(f)  { if( (f)!=(f) ){DBG_ASSERT(false);} }

#define DBG_ASSERT_MSG( val, errmsg )        \
	if(!(val)){                                \
	DBG_ASSERT( false )                        \
	}

#define DBG_ASSERT_VULKAN( val )             \
    DBG_ASSERT( VK_SUCCESS == val )

#define DBG_ASSERT_VULKAN_MSG( val, errmsg ) \
	if(!((VK_SUCCESS == val))){                \
    DBG_ASSERT( false )                      \
	}



/**************************************************************/
/*
   Helper function for loading shaders
*/
/**************************************************************/

inline
bool LoadShader(VkDevice          device,
                const char*       shaderFileName,
                VkShaderModule*   outShaderModule)
{
  // Simple Shaders Loader
  uint32_t codeSize = 0;
  // assume our file is less than 10,000 bytes
  char *code        = new char[10000];
  FILE* fileHandle  = NULL;

  // load your shader:
  fileHandle = fopen( shaderFileName, "rb" );

  // did you successfully find file
  DBG_ASSERT_MSG(fileHandle != NULL,
    "Failed to open shader file.");
  
  // read the file contents
  codeSize = fread(code, 1, 10000, fileHandle);
  // close the file
  fclose(fileHandle);
  fileHandle = NULL;

  VkShaderModuleCreateInfo vertexShaderCreationInfo = {};
  vertexShaderCreationInfo.sType     = VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO;
  vertexShaderCreationInfo.codeSize  = codeSize;
  vertexShaderCreationInfo.pCode     = (uint32_t *)code;

  VkResult result = 
  vkCreateShaderModule( device, 
                        &vertexShaderCreationInfo, 
                        NULL, 
                        outShaderModule );
  DBG_ASSERT_VULKAN_MSG( result, 
    "Failed to create vertex shader module." );

  memset(code, 0, 10000);
  delete code;
  code = NULL;

  return true;
}// End LoadShader(..)



/**************************************************************/
/*
   Minimal vector class of 3 floats and overloaded math operators
*/
/**************************************************************/

class Vector3 
{	
public:
	float x, y, z;

	Vector3	(float x, float y, float z)			{this->x=x; this->y=y; this->z=z;}
	Vector3	()									{}

	void	operator+= (const Vector3 &v)		{x+=v.x;   y+=v.y;   z+=v.z;}
	Vector3 operator/  (const float &v)	  const	{return Vector3(x/v,y/v,z/v);}
	Vector3 operator-  (const Vector3 &v) const	{return Vector3(x-v.x,y-v.y,z-v.z);}
	Vector3 operator+  (const Vector3 &v) const {return Vector3(x+v.x,y+v.y,z+v.z);}
	Vector3 operator*  (const float &v)   const	{return Vector3(x*v,y*v,z*v);}
	Vector3 operator-  ()				  const	{return Vector3(-x,-y,-z);}

	static Vector3  Cross	(const Vector3 &vA, const Vector3& vB)	
	{
		return Vector3(vA.y*vB.z - vA.z*vB.y, vA.z*vB.x - vA.x*vB.z, vA.x*vB.y - vA.y*vB.x);
	}
	static float	Dot		(const Vector3 &vA, const Vector3& vB)	{return vA.x*vB.x + vA.y*vB.y + vA.z*vB.z;}
	static Vector3 Normalize(const Vector3& v)
	{
		const float l = v.Length();
		return Vector3(v.x/l,v.y/l,v.z/l);
	}
	float	Length()	 const					{return sqrt(x*x+y*y+z*z);}
	float	LengthSq()	 const					{return (x*x+y*y+z*z);}

	static  double Length(const Vector3& v)
	{
		return v.Length();
	}
	
}; // End Vector3(..)



/**************************************************************/
/*
   Minimal matrix class basic operators and overloaded math
   functions
*/
/**************************************************************/

class Matrix4
{
public:
	float m[16];

	Matrix4::Matrix4()									
	{
		// Don't do anything in here - as it can be expensive to 
		// keep initializing Matrices (i.e., large arrays or reading
		// data from files)
	}


	// Copy constructor
	Matrix4::Matrix4(const Matrix4& rhs)
	{
		for (int i=0; i<16; ++i)
		{
			m[i] = rhs.m[i];
		}// End for i
	}

	Matrix4::Matrix4(	float _00, float _01, float _02, float _03,
						float _10, float _11, float _12, float _13,
						float _20, float _21, float _22, float _23,
						float _30, float _31, float _32, float _33)
	{
		m[0]=_00;	m[4]=_01;	m[8] =_02;	m[12]=_03;
		m[1]=_10;	m[5]=_11;	m[9] =_12;	m[13]=_13;
		m[2]=_20;	m[6]=_21;	m[10]=_22;	m[14]=_23;
		m[3]=_30;   m[7]=_31;	m[11]=_32;	m[15]=_33;
	
		/* 
		// For row-order
	m[0] = _00;	m[4]  = _10;	m[8]  = _20;	m[12] = _30;
	m[1] = _01;	m[5]  = _11;	m[9]  = _21;	m[13] = _31;
	m[2] = _02;	m[6]  = _12;	m[10] = _22;	m[14] = _32;
	m[3] = _03;	m[7]  = _13;	m[11] = _23;	m[15] = _33;
		*/
	}

  Matrix4 operator*  (const float &v)   const	
  {
    Matrix4 ret;
    for (int i=0; i<16; ++i)
    {
      ret[i] = m[i]*v;
    }// End for i
    return ret;
  }// End operator*


  Matrix4 operator+  (const Matrix4& rhs) const
  {
    Matrix4 ret;
    for (int i=0; i<16; ++i)
    {
      ret[i] = m[i] + rhs.m[i];
    }// End for i
    return ret;
  }// End operator*


	float&			Matrix4::operator[]	(int index)			{ return m[index]; }

	const float&	Matrix4::operator[]	(int index) const	{ return m[index]; }


  float Get(int x, int y) const
  {
    return m[x*4+y];
  }


	void Set(int x, int y, float val)
	{
		DBG_ASSERT(false);
		m[x*4+y] = val;
	}


	static Matrix4 Matrix4::Identity()
	{
		static Matrix4 identity(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1);
		return identity;
	}

  static Matrix4 Transpose(const Matrix4& mat)
  {
    Matrix4 tmp;
    for(int i=0;i<4;i++) 
    {
      for(int j=0;j<4;j++) 
      {
        tmp[4*i+j] = mat[i+4*j];
      }// End for j
    }// End for i
    return tmp;
  }// End Transpose(..)


	static 
	Matrix4 Matrix4::SetTranslation(const Vector3& translation)
	{
		Matrix4 mat = Matrix4::Identity();
		mat.m[12] = translation.x;
		mat.m[13] = translation.y;
		mat.m[14] = translation.z;
		return mat;
	}

	static Matrix4 SetScale			(const Vector3& scale)
	{
		Matrix4 mat = Matrix4::Identity();
		mat.m[ 0] = scale.x;
		mat.m[ 5] = scale.y;
		mat.m[10] = scale.z;
		return mat;
	}

	// ref: https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle
	static Matrix4 Matrix4::SetRotationAxis(float angle, const Vector3& axis)
	{
		// Page 466, Graphics Gems

		float s = sin(angle);
		float c = cos(angle);
		float t = 1 - c;

		Vector3 ax = Vector3::Normalize( axis );

		float x = ax.x;
		float y = ax.y;
		float z = ax.z;

		Matrix4 rotate(t*x*x+c, t*x*y-s*z, t*x*z+s*y, 0,
						t*y*x+s*z, t*y*y+c, t*y*z-s*x, 0,
						t*z*x-s*y, t*z*y+s*x, t*z*z+c, 0,
						0,			0,			0,		1 );

		return rotate;
	}

	// https://msdn.microsoft.com/en-us/library/windows/desktop/bb205342(v=vs.85).aspx
	// Left Handed
	//2*zn/w  0       0              0
	//0       2*zn/h  0              0
	//0       0       zf/(zf-zn)     1
	//0       0       zn*zf/(zn-zf)  0

	// https://msdn.microsoft.com/en-us/library/windows/desktop/bb205355(v=vs.85).aspx
	// Right Handed
	//2*zn/w  0       0              0
	//0       2*zn/h  0              0
	//0       0       zf/(zn-zf)    -1
	//0       0       zn*zf/(zn-zf)  0

	static 
	Matrix4 SetPerspectiveRH(float FOV, float Aspect, float ZNear, float ZFar)
	{
		float Width =    1.0f / tanf(FOV/2.0f);
		float Height = Aspect / tanf(FOV/2.0f);

		Matrix4 res;
		res[ 0] = Width;
		res[ 4] = 0.0f;
		res[ 8] = 0.0f;
		res[12] = 0.0f;

		res[ 1] = 0.0f;
		res[ 5] = -Height;
		res[ 9] = 0.0f;
		res[13] = 0.0f;
	
		res[ 2] = 0.0f;
		res[ 6] = 0.0f;
		res[10] = ZFar / (ZNear - ZFar);
		res[14] = ZFar * ZNear / (ZNear - ZFar);
	
		res[ 3] = 0.0f;
		res[ 7] = 0.0f;
		res[11] = -1.0f;
		res[15] = 0.0f;
		return res;
	}

	// https://msdn.microsoft.com/en-us/library/windows/desktop/bb205342(v=vs.85).aspx
	// Left Handed
	// zaxis = normal(at - eye)
	// xaxis = normal(cross(Up, zaxis))
	// yaxis = cross(zaxis, xaxis)
	// xaxis.x           yaxis.x           zaxis.x          0
	// xaxis.y           yaxis.y           zaxis.y          0
	// xaxis.z           yaxis.z           zaxis.z          0
	//-dot(xaxis, eye)  -dot(yaxis, eye)  -dot(zaxis, eye)  1

	// https://msdn.microsoft.com/en-us/library/windows/desktop/bb205343(v=vs.85).aspx
	// Right Handed
	//zaxis = normal(eye - at)
	//xaxis = normal(cross(Up, zaxis))
	//yaxis = cross(zaxis, xaxis)
	//xaxis.x           yaxis.x           zaxis.x          0
	//xaxis.y           yaxis.y           zaxis.y          0
	//xaxis.z           yaxis.z           zaxis.z          0
	//dot(xaxis, eye)   dot(yaxis, eye)   dot(zaxis, eye)  1

	static 
	Matrix4 SetLookAtRH(const Vector3 &eye, const Vector3 &at, const Vector3 &up)
	{
		Vector3 XAxis, YAxis, ZAxis;
		ZAxis = Vector3::Normalize(eye - at);
		XAxis = Vector3::Normalize(Vector3::Cross(up, ZAxis));
		YAxis = Vector3::Normalize(Vector3::Cross(ZAxis, XAxis));

		Matrix4 res;
		res[ 0] = XAxis.x;
		res[ 4] = XAxis.y;
		res[ 8] = XAxis.z;
		res[12] = -Vector3::Dot(XAxis,eye);
		
		res[ 1] = YAxis.x;
		res[ 5] = YAxis.y;
		res[ 9] = YAxis.z;
		res[13] = -Vector3::Dot(YAxis,eye);
		
		res[ 2] = ZAxis.x;
		res[ 6] = ZAxis.y;
		res[10] = ZAxis.z;
		res[14] = -Vector3::Dot(ZAxis,eye);
		
		res[ 3] = 0.0f;
		res[ 7] = 0.0f;
		res[11] = 0.0f;
		res[15] = 1.0f;
		return res;
	}

	// https://msdn.microsoft.com/en-us/library/windows/desktop/bb205349(v=vs.85).aspx
	// Right Handed
	//2/w  0    0           0
	//0    2/h  0           0
	//0    0    1/(zn-zf)   0
	//0    0    zn/(zn-zf)  1

	// https://msdn.microsoft.com/en-us/library/windows/desktop/bb205346(v=vs.85).aspx
	// Left Handed
	//2/w  0    0           0
	//0    2/h  0           0
	//0    0    1/(zf-zn)   0
	//0    0   -zn/(zf-zn)  1

	static Matrix4 SetOrthogonal(float Width, float Height, float ZNear, float ZFar)
	{
		float l = 0;
		float r = Width;
		float t = 0;
		float b = Height;

		Matrix4 res;
		res[ 0] = 2.0f / (r-l);
		res[ 1] = 0.0f;
		res[ 2] = 0.0f;
		res[ 3] = 0.0f;

		res[ 4] = 0.0f;
		res[ 5] = 2.0f / (t-b);
		res[ 6] = 0.0f;
		res[ 7] = 0.0f;

		res[ 8] = 0.0f;
		res[ 9] = 0.0f;
		res[10] = -2.0f / (ZFar - ZNear);
		res[11] = 0;

		res[12] = -(r+l)/(r-l);
		res[13] = -(t+b)/(t-b);
		res[14] = -(ZFar+ZNear)/(ZFar-ZNear);
		res[15] = 1.0f;
		return res;
	}
	
}; // End Matrix4(..)




inline
Matrix4 operator*(const Matrix4& lhs, const Matrix4& rhs)
{
	Matrix4 result;

	result[0] = lhs[0]*rhs[0] + lhs[4]*rhs[1] + lhs[8]*rhs[2] + lhs[12]*rhs[3];
	result[1] = lhs[1]*rhs[0] + lhs[5]*rhs[1] + lhs[9]*rhs[2] + lhs[13]*rhs[3];
	result[2] = lhs[2]*rhs[0] + lhs[6]*rhs[1] + lhs[10]*rhs[2] + lhs[14]*rhs[3];
	result[3] = lhs[3]*rhs[0] + lhs[7]*rhs[1] + lhs[11]*rhs[2] + lhs[15]*rhs[3];
	
	result[4] = lhs[0]*rhs[4] + lhs[4]*rhs[5] + lhs[8]*rhs[6] + lhs[12]*rhs[7];
	result[5] = lhs[1]*rhs[4] + lhs[5]*rhs[5] + lhs[9]*rhs[6] + lhs[13]*rhs[7];
	result[6] = lhs[2]*rhs[4] + lhs[6]*rhs[5] + lhs[10]*rhs[6] + lhs[14]*rhs[7];
	result[7] = lhs[3]*rhs[4] + lhs[7]*rhs[5] + lhs[11]*rhs[6] + lhs[15]*rhs[7];

	result[8] = lhs[0]*rhs[8] + lhs[4]*rhs[9] + lhs[8]*rhs[10] + lhs[12]*rhs[11];
	result[9] = lhs[1]*rhs[8] + lhs[5]*rhs[9] + lhs[9]*rhs[10] + lhs[13]*rhs[11];
	result[10] = lhs[2]*rhs[8] + lhs[6]*rhs[9] + lhs[10]*rhs[10] + lhs[14]*rhs[11];
	result[11] = lhs[3]*rhs[8] + lhs[7]*rhs[9] + lhs[11]*rhs[10] + lhs[15]*rhs[11];

	result[12] = lhs[0]*rhs[12] + lhs[4]*rhs[13] + lhs[8]*rhs[14] + lhs[12]*rhs[15];
	result[13] = lhs[1]*rhs[12] + lhs[5]*rhs[13] + lhs[9]*rhs[14] + lhs[13]*rhs[15];
	result[14] = lhs[2]*rhs[12] + lhs[6]*rhs[13] + lhs[10]*rhs[14] + lhs[14]*rhs[15];
	result[15] = lhs[3]*rhs[12] + lhs[7]*rhs[13] + lhs[11]*rhs[14] + lhs[15]*rhs[15];

	return result;
}// End operator*




inline
Vector3 Transform(const Matrix4& mat, const Vector3& p)
{
  return Vector3(
  mat[0]*p.x +
  mat[4]*p.y +
  mat[8]*p.z,

  mat[1]*p.x +
  mat[5]*p.y +
  mat[9]*p.z,

  mat[2]*p.x +
  mat[6]*p.y +
  mat[10]*p.z
  ) + Vector3( mat[12], mat[13], mat[14] );
}// End Transform(..)



inline
Vector3 operator*(const Matrix4& lhs, const Vector3& rhs)
{
#if 0
	Vector3 result;

	result.x = lhs[0]*rhs.x + lhs[4]*rhs.y + lhs[8] *rhs.z + lhs[12];
	result.y = lhs[1]*rhs.x + lhs[5]*rhs.y + lhs[9] *rhs.z + lhs[13];
	result.z = lhs[2]*rhs.x + lhs[6]*rhs.y + lhs[10]*rhs.z + lhs[14];
	//result[3] = lhs[3]*rhs[0] + lhs[7]*rhs[1] + lhs[11]*rhs[2] + lhs[15]*rhs[3];
  return result;
#else

  return Transform( lhs, rhs);

#endif

}// End operator*



