/****************************************************************/
/*  
     ***   ***  ***  *** ***     *** ***   ******  ****  ***
     ***   ***  ***  *** ***     *** **   ***  *** ***** ***
     ***   ***  ***  *** ***     *****    ******** *** *****
      *** ***   ***  *** ******* *** **   ***  *** ***  ****
       *****     *****   ******* *** ***  ***  *** ***   ***

	 Vulkan Graphics API
	 Introductory Tutorials

   Minimal working examples (MWE) to build upon and 
   modify to gain an understanding of the API

/*
/****************************************************************/


/**************************************************************/
/*
   skin.h

   About
   Simple skinning demo - generates a tubular mesh and a linked
   chain of bones.  Simple distance calculation the bones are 
   connected to the mesh through weights.

   Provides a simple self contained example on the topic of 
   skinning and blended deformation that is easy to
   reverse engineer and expand upon
*/
/**************************************************************/

#pragma once

#include <vector>
using namespace std;

#include "helpers.h" // simple vector/matrix classes


#include "vulkan.h"

#include "debugmesh.h"


/**************************************************************/
/*
   
*/
/**************************************************************/

#define VERTEX_BUFFER_BIND_ID   0

// Maximum number of bones
#define MAX_BONES               50

// Maximum weights per vertex
#define MAX_JOINT_WEIGHTS       4  


class Skin
{
public:
  struct stVertex 
  {
    Vector3   pos;
    Vector3   normal;

    // maximum of MAX_JOINT_WEIGHTS bones influence
    uint32_t  boneIndex[MAX_JOINT_WEIGHTS];
    float     boneWeight[MAX_JOINT_WEIGHTS];
  };


  // Structure must match the structure defined uniform in shader
  struct stUniform 
  {
    Matrix4 model;
    Matrix4 view;
    Matrix4 projection;
    Matrix4 bones[MAX_BONES];
  };


  struct stBone
  {
    float   length;   // translation along local y-axis
    Vector3 localRot;
    Matrix4 worldTranform;
    Matrix4 invWorldRestTransform;
    Matrix4 skinWorldTransform;
  };


  struct stVerticeDescription
  {
    VkPipelineVertexInputStateCreateInfo           inputState;
    std::vector<VkVertexInputBindingDescription>   bindingDescriptions;
    std::vector<VkVertexInputAttributeDescription> attributeDescriptions;
  };

  VkBuffer			        vbuf;
  VkDeviceMemory        vmem;
  size_t				        vsize;

  VkBuffer			        ibuf;
  VkDeviceMemory	      imem;
  size_t				        isize;

  VkBuffer			        ubuf;
  VkDeviceMemory	      umem;
  size_t				        usize;

  uint32_t              indexCount;
  Vector3	              dim;
  stVerticeDescription  verticeDescription;

  vector<stBone>        bones;




	Skin( VkDevice         device,
        VkPhysicalDevice physicalDevice )
	{
    vbuf               = VK_NULL_HANDLE;
    vmem               = VK_NULL_HANDLE;
    vsize              = 0;
	  
    ibuf               = VK_NULL_HANDLE;
    imem               = VK_NULL_HANDLE;
    isize              = 0;

    ubuf               = VK_NULL_HANDLE;
    umem               = VK_NULL_HANDLE;
    usize              = 0;
    
    indexCount         = 0;
    dim                = Vector3(0,0,0);
    memset(&verticeDescription, 0, sizeof(verticeDescription));


    CreateBonesAndMesh(device, physicalDevice);

    SetupVertexDescriptions();

  }// Skin(..)


  void SetupVertexDescriptions()
  {
    // Binding description
    verticeDescription.bindingDescriptions.resize(1);

    VkVertexInputBindingDescription vInputBindDescription = {};
    vInputBindDescription.binding                        = VERTEX_BUFFER_BIND_ID;
    vInputBindDescription.stride                         = sizeof(stVertex);
    vInputBindDescription.inputRate                      = VK_VERTEX_INPUT_RATE_VERTEX;
    verticeDescription.bindingDescriptions[0]            = vInputBindDescription;

    // Attribute descriptions
    verticeDescription.attributeDescriptions.resize(4);

    // Location 0 : Position
    verticeDescription.attributeDescriptions[0].location = 0;
    verticeDescription.attributeDescriptions[0].binding  = VERTEX_BUFFER_BIND_ID;
    verticeDescription.attributeDescriptions[0].format   = VK_FORMAT_R32G32B32_SFLOAT;
    verticeDescription.attributeDescriptions[0].offset   = 0;

    // Location 2 : Normal
    verticeDescription.attributeDescriptions[1].location = 1;
    verticeDescription.attributeDescriptions[1].binding  = VERTEX_BUFFER_BIND_ID;
    verticeDescription.attributeDescriptions[1].format   = VK_FORMAT_R32G32B32_SFLOAT;
    verticeDescription.attributeDescriptions[1].offset   = sizeof(float) * 3;

    // Location 3 : Bone Index
    verticeDescription.attributeDescriptions[2].location = 2;
    verticeDescription.attributeDescriptions[2].binding  = VERTEX_BUFFER_BIND_ID;
    verticeDescription.attributeDescriptions[2].format   = VK_FORMAT_R32G32B32A32_SINT;
    verticeDescription.attributeDescriptions[2].offset   = sizeof(float) * 6;

    // Location 4 : Bone Weight
    verticeDescription.attributeDescriptions[3].location = 3;
    verticeDescription.attributeDescriptions[3].binding  = VERTEX_BUFFER_BIND_ID;
    verticeDescription.attributeDescriptions[3].format   = VK_FORMAT_R32G32B32A32_SFLOAT;
    verticeDescription.attributeDescriptions[3].offset   = sizeof(float) * 10;


    VkPipelineVertexInputStateCreateInfo pipelineVertexInputStateCreateInfo = {};
    pipelineVertexInputStateCreateInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO;
    pipelineVertexInputStateCreateInfo.pNext = NULL;

    verticeDescription.inputState                                 = pipelineVertexInputStateCreateInfo;
    verticeDescription.inputState.vertexBindingDescriptionCount   = verticeDescription.bindingDescriptions.size();
    verticeDescription.inputState.pVertexBindingDescriptions      = verticeDescription.bindingDescriptions.data();
    verticeDescription.inputState.vertexAttributeDescriptionCount = verticeDescription.attributeDescriptions.size();
    verticeDescription.inputState.pVertexAttributeDescriptions    = verticeDescription.attributeDescriptions.data();
  }// End SetupVertexDescriptions(..)


	void CreateBonesAndMesh(VkDevice         device,
                          VkPhysicalDevice physicalDevice)
	{
    // Number of single connected links for this example
    const int numberOfLinks = 3;
    // Length of each link
    float length = 1.0f;

    // Initialize array of bones
    for (int i=0; i<numberOfLinks; ++i)
    {
      stBone bone;
      bone.length                = length;
      bone.localRot              = Vector3(0,0,0);
      bone.worldTranform         = Matrix4::Identity();
      bone.invWorldRestTransform = Matrix4::Identity();
      bone.skinWorldTransform    = Matrix4::Identity();
      bones.push_back( bone );
    }// End for i


    // Construct single large mesh (long tube the same length as the
    // linked bones)
#if 1
    // resolution of your tubular mesh
    int numCircleSegs = 10;
    int numLengthSegs = 30;

    // radius of the tube
    float radius = 0.1f;

    // Part 1 of the mesh generation - build up the
    // rings of the vertices along the y-axis

    // All the vertices for the mesh
    vector<stVertex> vertexBuffer;

    for (int h=0; h<numLengthSegs; ++h)
    {
      const float hn = h / (float)(numLengthSegs-1);

      for (int c=0; c<numCircleSegs; ++c)
      {
        const float cn = c / (float)(numCircleSegs-2);
      
        const float x  = sinf(cn * 3.1415f * 2.0f) * radius;
        const float z  = cosf(cn * 3.1415f * 2.0f) * radius;
        const float y  = hn * length * bones.size();
      
        stVertex v;
        memset(&v, 0, sizeof(stVertex));
        v.pos    = Vector3(x, y, z);
        v.normal = Vector3::Normalize( Vector3(x, 0, z) );

        vertexBuffer.push_back( v );
      }// End for c
    }// End for h


    // Part 2 of the mesh generation - construct the triangle
    // indices that make up the surface of the tubular mesh
    vector<uint32_t> indexBuffer;

    const int numVertices = numLengthSegs * numCircleSegs;
    DBG_ASSERT(numVertices==vertexBuffer.size());

    int ii = 0;
    for (int h=0; h<numLengthSegs-1; ++h)
    {
      for (int c=0; c<numCircleSegs-2; ++c)
      {
          int i0 = ii                    ;
          int i1 = ii + 1                ;
          int i2 = ii + 1 + numCircleSegs;
          int i3 = ii +     numCircleSegs;


          indexBuffer.push_back( i0 );
          indexBuffer.push_back( i1 );
          indexBuffer.push_back( i3 );

          indexBuffer.push_back( i1 );
          indexBuffer.push_back( i2 );
          indexBuffer.push_back( i3 );

          DBG_ASSERT(indexBuffer[indexBuffer.size()-1] < vertexBuffer.size() );
          DBG_ASSERT(indexBuffer[indexBuffer.size()-2] < vertexBuffer.size() );
          DBG_ASSERT(indexBuffer[indexBuffer.size()-3] < vertexBuffer.size() );
          DBG_ASSERT(indexBuffer[indexBuffer.size()-4] < vertexBuffer.size() );
          DBG_ASSERT(indexBuffer[indexBuffer.size()-5] < vertexBuffer.size() );
          DBG_ASSERT(indexBuffer[indexBuffer.size()-6] < vertexBuffer.size() );

          ii++;
      }// End for c
      ii+=2;
    }// End for h



    // Determine which bones are connected to which vertices
    // Simple distance check
#if 1
    struct TmpBoneWeights 
    {
      vector<int>   boneIndex;
      vector<float> boneWeight;
    };
    vector<TmpBoneWeights> tmpBoneWeights;
    tmpBoneWeights.resize(vertexBuffer.size());

    // connect bones to vertices - simple distance check
    for (int a=0; a<(int)vertexBuffer.size(); ++a)
    {
      const Vector3& vp = vertexBuffer[a].pos;

      bool found = false;
      for (int i=0; i<(int)bones.size(); ++i)
      {
        float length = bones[i].length;
        Vector3 boneMiddle = Vector3(0, i+length*0.5f, 0);
        

        // Distance along the bone (vertically) - assume a 
        // singly linked chain of bones
        float dist = abs( vp.y - boneMiddle.y );

        if ( dist < 0.75f )
        {
          found |= true;

          bool add = true;
          for (int k=0; k<(int)tmpBoneWeights[a].boneIndex.size(); ++k)
          {
            if ( tmpBoneWeights[a].boneIndex[k] == i )
            {
              add = false;
            }
          }
          if ( add )
          {
            tmpBoneWeights[a].boneIndex.push_back( i );
            DBG_ASSERT( dist > 0.01f );
            tmpBoneWeights[a].boneWeight.push_back( 0.75f - dist );// 1.0 / dist );// 1 / dist );
          }
          DBG_ASSERT(tmpBoneWeights[a].boneIndex.size()<3);

        }// End if dist

      }// End for i

      // At least one bone for each vertex
      DBG_ASSERT(found);
    }// End for a

    // Normalize the weights - the weights should sum up to 1.0
    // Loop over every vertex, sum the weight - then divide by the sum
    // to get a normalized value between 0 and 1.0
    DBG_ASSERT( tmpBoneWeights.size() == vertexBuffer.size() );
    for (int i=0; i<(int)vertexBuffer.size(); ++i)
    {
      DBG_ASSERT(tmpBoneWeights[i].boneIndex.size()<=2);
      DBG_ASSERT(tmpBoneWeights[i].boneWeight.size() == tmpBoneWeights[i].boneIndex.size());

      float totalWeightt = 0;
      for (int g=0; g<(int)tmpBoneWeights[i].boneWeight.size(); ++g)
      {
        totalWeightt += tmpBoneWeights[i].boneWeight[g];
      }// End for g

      for (int g=0; g<(int)tmpBoneWeights[i].boneWeight.size(); ++g)
      {
        float tmp = tmpBoneWeights[i].boneWeight[g];
        tmp = tmp * (1.0f / totalWeightt);
        tmpBoneWeights[i].boneWeight[g] =  tmp;
      }// End for g

      // Copy the bone indices and weights back to the original vertex structure
      for (int g=0; g<(int)tmpBoneWeights[i].boneWeight.size(); ++g)
      {
        vertexBuffer[i].boneIndex[g]  = tmpBoneWeights[i].boneIndex[g];
        vertexBuffer[i].boneWeight[g] = tmpBoneWeights[i].boneWeight[g];
      }// End for g

    }// End for i
#endif

#endif

    // Calculate rest inverse transforms for each bone
    for (int i=0; i<(int)bones.size(); ++i)
    {
      // no rotation and the rest pose of joints is along
      // the vertical axis (hence the inverse is simply the
      // negation of the y-translation)

      // inverse
      float length = bones[i].length;
      bones[i].invWorldRestTransform = Matrix4::SetTranslation( Vector3(0, -length*(i+1), 0) );

    }// End for i




    // Create Vertex Buffer
    CreateBuffer( device,
                  physicalDevice,
                  VK_BUFFER_USAGE_VERTEX_BUFFER_BIT,
                  VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT,
                  sizeof(stVertex) * vertexBuffer.size(),
                  &vertexBuffer[0],
                  &vbuf,
                  &vmem);

    // Setup indices
    indexCount = indexBuffer.size();

    // Create index buffer
    CreateBuffer( device,
                  physicalDevice,
                  VK_BUFFER_USAGE_INDEX_BUFFER_BIT,
                  VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT,
                  sizeof(uint32_t) * indexBuffer.size(),
                  &indexBuffer[0],
                  &ibuf,
                  &imem);

    // Create uniform buffer
    CreateBuffer( device,
                  physicalDevice,
                  VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT,
                  VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT,
                  sizeof(stUniform),
                  nullptr,
                  &ubuf,
                  &umem);

	}


  void UpdateBoneTransforms()
  {
    // Bone Hierarchy
    for (int i=0; i<(int)bones.size(); ++i)
    {
      stBone& b = bones[i];

      Matrix4 t = Matrix4::SetTranslation( Vector3(0, b.length, 0) );
      Matrix4 r = Matrix4::SetRotationAxis( b.localRot.x, Vector3(1,0,0) ) *
                  Matrix4::SetRotationAxis( b.localRot.y, Vector3(0,1,0) ) *
                  Matrix4::SetRotationAxis( b.localRot.z, Vector3(0,0,1) );

      
      Matrix4 parentTransform = Matrix4::Identity();
      if ( i > 0 )
      {
        parentTransform = bones[i-1].worldTranform;
      }// End if i


      b.worldTranform = parentTransform * (r*t);

      Vector3 p0 = Vector3(parentTransform[12], parentTransform[13], parentTransform[14]);
      Vector3 p1 = Vector3(b.worldTranform[12], b.worldTranform[13], b.worldTranform[14]);
      DebugMesh::AddDrawLine(p0, p1, Vector3(0.2f,0,0.2f) );

    }// End for i
    DebugMesh::AddDrawCross(Vector3(0,1,0), 0.1f, Vector3(0.2f,0,0.0f) );

    // Skinning transforms
    for (int i=0; i<(int)bones.size(); ++i)
    {
      stBone& b = bones[i];
    
      bones[i].skinWorldTransform = bones[i].worldTranform * bones[i].invWorldRestTransform;
    }// End for i
    
  }// End UpdateBoneTransforms(..)







  void UpdateCommandBuffer(VkDevice         device,
                           VkPhysicalDevice physicalDevice,
                           uint32_t         nextImageIdx,
                           VkFramebuffer*   frameBuffers,
                           int              width,
                           int              height,
                           VkCommandBuffer  commandBuffer,
                           VkRenderPass     renderPass,
                           VkDescriptorSet  descriptorSet,
                           VkPipelineLayout pipelineLayout,
                           VkPipeline       pipeline)
  {
    
  // Build the command buffer

  VkCommandBufferBeginInfo beginInfo            = {};
	beginInfo.sType                               = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
  beginInfo.flags                               = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT;
	beginInfo.pNext                               = NULL;

  vkBeginCommandBuffer(commandBuffer, 
                       &beginInfo );

  // activate render pass:
  VkClearValue clearValue[]                     = { { 1.0f, 1.0f, 1.0f, 1.0f }, { 1.0, 0.0 } };
  VkRenderPassBeginInfo renderPassBeginInfo     = {};
  renderPassBeginInfo.sType                     = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO;
  renderPassBeginInfo.renderPass                = renderPass;
  renderPassBeginInfo.framebuffer               = frameBuffers[ nextImageIdx ];
  VkOffset2D a                                  = {0, 0}; 
  VkExtent2D b                                  = {width, height};
  VkRect2D   c                                  = {a,b}; 
  renderPassBeginInfo.renderArea                = c;
  renderPassBeginInfo.clearValueCount           = 2;
  renderPassBeginInfo.pClearValues              = clearValue;
  vkCmdBeginRenderPass( commandBuffer, 
                        &renderPassBeginInfo, 
                        VK_SUBPASS_CONTENTS_INLINE );

  // bind the graphics pipeline to the command buffer. 
  // Any vkDraw command afterwards is affected by this
  // pipeline.
  vkCmdBindPipeline( commandBuffer, 
                     VK_PIPELINE_BIND_POINT_GRAPHICS, 
                     pipeline );    

  // take care of dynamic state:
  VkViewport viewport = { 0, 0, (float)width, (float)height, 0, 1 };
  vkCmdSetViewport( commandBuffer, 0, 1, &viewport );

  VkRect2D scissor = { 0, 0, width, height };
  vkCmdSetScissor( commandBuffer, 0, 1, &scissor);

  // bind our shader parameters:
  vkCmdBindDescriptorSets( commandBuffer, 
                           VK_PIPELINE_BIND_POINT_GRAPHICS, 
                           pipelineLayout, 
                           0, 
                           1, 
                           &descriptorSet, 
                           0, 
                           NULL );


  VkDeviceSize offsets = { 0 };
	vkCmdBindVertexBuffers(commandBuffer, VERTEX_BUFFER_BIND_ID, 1, &vbuf, &offsets);
	vkCmdBindIndexBuffer(commandBuffer, ibuf, 0, VK_INDEX_TYPE_UINT32);
  vkCmdDrawIndexed(commandBuffer, indexCount, 1, 0, 0, 0);


  vkCmdEndRenderPass( commandBuffer );

  vkEndCommandBuffer( commandBuffer );

  static float count = 0.0f;
  count += 0.001f;

  // testing - slowly rotate one joint one way and another joint another way
  bones[1].localRot.z = sinf(count) * 0.5f;

  bones[2].localRot.z = sinf(count*1.5f) * 0.4f;

  }// End UpdateCommandBuffer()
  



  void UpdateUniformBuffer (VkDevice          device,
                            VkPhysicalDevice  physicalDevice,
                            int               width,
                            int               height,
                            VkDeviceMemory    memory)
  {
    static
    Vector3             rotation       = Vector3(0,  0,     0);
    Vector3             cameraPos      = Vector3(0,  1.5f, -1.5f);
    Vector3             mousePos       = Vector3(0,  0,     0);
    float               zoom           = 1;

    // slowly rotate around the object you're looking at
    //rotation.y += 0.0001f;

    Matrix4 m = Matrix4::SetRotationAxis(rotation.y, Vector3(0,1,0)) *
                Matrix4::SetRotationAxis(rotation.x, Vector3(1,0,0));

    Matrix4 p = Matrix4::SetPerspectiveRH( 2, // radians not degrees
                                          (float)width / (float)height,
                                          0.2f, 
                                          1000.0f);

    Matrix4 v = Matrix4::SetLookAtRH( cameraPos * zoom, // Always looking at origin
                                      Vector3(0,1.5f,0),
                                      Vector3(0,1,0));

    // Skin uniforms
    stUniform ubo;
    ubo.view       = (v);
    ubo.projection = (p);
    ubo.model      = (m);

    for (int i=0; i<(int)bones.size(); ++i)
    {
      ubo.bones[i] = (bones[i].skinWorldTransform);
    }// End for i

    uint8_t *pData;
    DBG_ASSERT_VULKAN(vkMapMemory(device, memory, 0, sizeof(ubo), 0, (void **)&pData));
    memcpy(pData, &ubo, sizeof(ubo));
    vkUnmapMemory(device, memory);

    DebugMesh::SetModelViewProjection(m,v,p);

  }// End UpdateUniformBuffers(..)





  static
  VkBool32 CreateBuffer(VkDevice              device,
                        VkPhysicalDevice      physicalDevice,
                        VkBufferUsageFlags    usageFlags, 
                        VkMemoryPropertyFlags memoryPropertyFlags, 
                        VkDeviceSize          size, 
                        void *                data, 
                        VkBuffer *            buffer, 
                        VkDeviceMemory *      memory)
  {
    VkMemoryRequirements memReqs;
    VkMemoryAllocateInfo memAlloc       = {};
    memAlloc.sType                      = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
    memAlloc.pNext                      = NULL;
    memAlloc.allocationSize             = 0;
    memAlloc.memoryTypeIndex            = 0;

    VkBufferCreateInfo bufferCreateInfo = {};
    bufferCreateInfo.sType              = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
    bufferCreateInfo.pNext              = NULL;
    bufferCreateInfo.usage              = usageFlags;
    bufferCreateInfo.size               = size;
    bufferCreateInfo.flags              = 0;

    DBG_ASSERT_VULKAN(vkCreateBuffer(device, &bufferCreateInfo, nullptr, buffer));

    vkGetBufferMemoryRequirements(device, *buffer, &memReqs);
    memAlloc.allocationSize  = memReqs.size;
    memAlloc.memoryTypeIndex = GetMemoryType(physicalDevice, memReqs.memoryTypeBits, memoryPropertyFlags);
    DBG_ASSERT_VULKAN(vkAllocateMemory(device, &memAlloc, nullptr, memory));
    if (data != nullptr)
    {
      void *mapped;
      DBG_ASSERT_VULKAN(vkMapMemory(device, *memory, 0, size, 0, &mapped));
      memcpy(mapped, data, (size_t)size);
      vkUnmapMemory(device, *memory);
    }
    DBG_ASSERT_VULKAN(vkBindBufferMemory(device, *buffer, *memory, 0));

    return true;
  }// End CreateBuffer(..)


  static
  uint32_t GetMemoryType(VkPhysicalDevice      physicalDevice,
                         uint32_t              typeBits,
                         VkMemoryPropertyFlags properties)
  {
    // read the device memory properties
    VkPhysicalDeviceMemoryProperties memoryProperties;
    vkGetPhysicalDeviceMemoryProperties( physicalDevice, &memoryProperties );

    bool memTypeFound = false;
    for (uint32_t i = 0; i < memoryProperties.memoryTypeCount; i++)
    {
      if ((typeBits & 1) == 1)
      {
        if ((memoryProperties.memoryTypes[i].propertyFlags & properties) == properties)
        {
          if (memTypeFound)
          {
            memTypeFound = true;
          }
          return i;
        }
      }
      typeBits >>= 1;
    }

    if (!memTypeFound)
    {
      __debugbreak(); // Could not find a matching memory type
      return 0;
    }// End if
    return 1;
  }// End GetMemoryType(..)


}; // End class Skin




